"""The command line interface implementation"""
import argparse
import configparser
import os
import shlex
import shutil
import stat
import sys
import tempfile
from collections import defaultdict
from functools import partial
from pathlib import Path
from shutil import which

from rich.console import Console

from frysk.process import execute
from frysk.settings import (
    merge_settings,
    settings_from,
)
from frysk.test import runtests
from frysk.xunit import runxunit

VERSION = "0.15.1"


class ExitCode:
    """Possible exit codes of frysk CLI"""

    SUCCESS = 0
    TEST_FAILED = 1
    ERROR = 2


def main(argv=None):
    """Main entry point.

    If you're thinking of using frysk in other Python code (e.g., unit tests),
    consider using the test() or testfile() functions instead.

    :param argv: Script arguments (excluding script name)
    :type argv: iterable of strings
    :return: Exit code (non-zero on failure)
    :rtype: int
    """
    return _Cli().main(argv)


def load(config, supported, section="frysk"):
    """
    Load configuration options from a init style format config file.

    :param supported: iterable of supported options and their type which should be collected.
    :param section: which contains the options.
    """
    parser = configparser.ConfigParser()
    parser.read(config)
    dispatcher = defaultdict(
        lambda: (parser.get, "--{}: invalid value: {!r}"),
        {
            bool: (parser.getboolean, "--{}: invalid boolean value: {!r}"),
            int: (parser.getint, "--{}: invalid integer value: {!r}"),
        },
    )
    if not parser.has_section(section):
        return {}

    config = {}
    for _type, option in supported:
        if not parser.has_option(section, option):
            continue
        try:
            fetch, error_msg = dispatcher[_type]
            config[option] = fetch(section, option)
        except ValueError as ex:
            fetch, error_msg = dispatcher[_type]
            value = parser.get(section, option)
            raise ValueError(error_msg.format(option, value)) from ex
    return config


def _conflicts(settings):
    conflicts = [
        ("--yes", settings.yes, "--no", settings.no),
        ("--quiet", settings.quiet, "--interactive", settings.interactive),
        ("--debug", settings.debug, "--quiet", settings.quiet),
        ("--debug", settings.debug, "--interactive", settings.interactive),
        ("--debug", settings.debug, "--verbose", settings.verbose),
        ("--debug", settings.debug, "--xunit-file", settings.xunit_file),
    ]
    for option1, value1, option2, value2 in conflicts:
        if value1 and value2:
            return option1, option2
    return None


def _env_args(var, env=None):
    env = env if env else os.environ
    args = env.get(var, "").strip()
    return shlex.split(args)


class _ArgumentParser:
    """argparse.Argumentparser compatible argument parser.

    Allows inspection of options supported by the parser"""

    @classmethod
    def create_parser(cls):
        parser = cls(
            usage="frysk [OPTIONS] TESTS...",
            prog="frysk",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
        parser.add_argument(
            "tests",
            metavar="TESTS",
            type=Path,
            nargs="+",
            help="Path(s) to the tests to be executed",
        )
        parser.add_argument("-V", "--version", action="version", version=VERSION)
        parser.add_argument(
            "-q", "--quiet", action="store_true", help="don't print diffs"
        )
        parser.add_argument(
            "-v",
            "--verbose",
            action="store_true",
            help="show filenames and test status",
        )
        parser.add_argument(
            "-i",
            "--interactive",
            action="store_true",
            help="interactively merge changed test output",
        )
        parser.add_argument(
            "-d",
            "--debug",
            action="store_true",
            help="write script output directly to the terminal",
        )
        parser.add_argument(
            "-y", "--yes", action="store_true", help="answer yes to all questions"
        )
        parser.add_argument(
            "-n", "--no", action="store_true", help="answer no to all questions"
        )
        parser.add_argument(
            "-E",
            "--preserve-env",
            action="store_true",
            help="don't reset common environment variables",
        )
        parser.add_argument(
            "--keep-tmpdir",
            action="store_true",
            help="keep temporary directories",
        )
        parser.add_argument(
            "--shell",
            action="store",
            default="/bin/sh",
            metavar="PATH",
            help="shell to use for running tests",
        )
        parser.add_argument(
            "--shell-opts",
            action="store",
            metavar="OPTS",
            help="arguments to invoke shell with",
        )
        parser.add_argument(
            "--indent",
            action="store",
            default=2,
            metavar="NUM",
            type=int,
            help="number of spaces to use for indentation",
        )
        parser.add_argument(
            "--color",
            choices=["always", "never", "auto"],
            default="auto",
            help="Mode which shall be used for coloring the output",
        )
        parser.add_argument(
            "--xunit-file",
            action="store",
            metavar="PATH",
            help="path to write xUnit XML output",
        )
        parser.add_argument(
            "--dos2unix",
            action="store_true",
            help="convert DOS/Windows line endings to UNIX line endings",
        )
        return parser

    def __init__(self, *args, **kwargs):
        self._options = []
        self._parser = argparse.ArgumentParser(*args, **kwargs)

    def add_argument(self, *args, **kwargs):
        """See argparser.Argumentparser:add_argument"""

        def is_boolean_option(a):
            return a.nargs is not None and isinstance(a.const, bool)

        action = self._parser.add_argument(*args, **kwargs)
        if not action.type:
            _type = bool if is_boolean_option(action) else None
        else:
            _type = action.type
        self._options.append((_type, action.dest))
        return action

    def __getattr__(self, item):
        return getattr(self._parser, item)

    @property
    def options(self):
        """
        Normalized options and their type except for -V, --version and -h, --help.

        :return: an iterable containing all boolean options.
        :rtype: Iterable[Tuple(type, str)]
        """
        return self._options


class _CliError(Exception):
    def __init__(self, exit_code, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._exit_code = exit_code

    @property
    def exit_code(self):
        return self._exit_code


class _Cli:
    @staticmethod
    def _expandpath(path):
        """Expands ~ and environment variables in path"""
        return os.path.expanduser(os.path.expandvars(path))

    @staticmethod
    def _patch(cmd, diff):
        """Run echo [lines from diff] | cmd -p0"""
        _, retcode = execute([cmd, "-p0"], stdin=b"".join(diff))
        return retcode == 0

    def __init__(self):
        self._stdout_console = Console(file=sys.stdout)
        self._stderr_console = Console(file=sys.stderr)
        self._argparser = _ArgumentParser.create_parser()
        self._default_color_system = self._stdout_console.color_system
        self.tmpdir = None

    @property
    def stdout(self):
        return partial(
            self._stdout_console.print, no_wrap=True, overflow="ignore", crop=False
        )

    @property
    def stderr(self):
        return partial(self._stderr_console.print, no_wrap=True)

    def _color_mode(self, mode):
        dispatcher = {"auto": "auto", "never": None, "always": "standard"}
        mode = dispatcher[mode]
        self._stdout_console = Console(file=sys.stdout, color_system=mode)
        self._stderr_console = Console(file=sys.stderr, color_system=mode)

    def _log(self, msg=None, verbosemsg=None, verbose=False):
        """Write msg to standard out and flush.

        If verbose is True, write verbosemsg instead.
        """
        msg = verbosemsg if verbose else msg
        if not msg:
            return
        msg = msg.encode("utf-8") if isinstance(msg, bytes) else msg
        self.stdout(msg, end="")

    def _runcli(self, tests, quiet=False, verbose=False, patchcmd=None, answer=None):
        """Run tests with command line interface input/output.

        tests should be a sequence of 2-tuples containing the following:

            (test path, test function)

        This function yields a new sequence where each test function is wrapped
        with a function that handles CLI input/output.

        If quiet is True, diffs aren't printed. If verbose is True,
        filenames and status information are printed.

        If patchcmd is set, a prompt is written to stdout asking if
        changed output should be merged back into the original test. The
        answer is read from stdin. If 'y', the test is patched using patch
        based on the changed output.
        """
        total, skipped, failed = [0], [0], [0]

        for path, test in tests:

            def testwrapper():
                """Test function that adds CLI output"""
                total[0] += 1
                self._log(None, f"{Path(path.parent.name, path.name)}: ", verbose)

                refout, postout, diff = test()
                if refout is None:
                    skipped[0] += 1
                    self._log("[yellow]s[/yellow]", "empty\n", verbose)
                    return refout, postout, diff

                errpath = Path(f"{path}" + ".err")
                if postout is None:
                    skipped[0] += 1
                    self._log(
                        "[yellow]s[/yellow]", "[yellow]skipped[/yellow]\n", verbose
                    )
                elif not diff:
                    self._log("[green].[/green]", "[green]passed[/green]\n", verbose)
                    if errpath.exists():
                        os.remove(errpath)
                else:
                    failed[0] += 1
                    self._log("[red]![/red]", "[red]failed[/red]\n", verbose)
                    if not quiet:
                        self._log("\n", None, verbose)

                    with open(errpath, "wb") as errfile:
                        for line in postout:
                            errfile.write(line)

                    if not quiet:
                        origdiff = diff
                        diff = []
                        for line in origdiff:
                            _line = line.decode("utf-8")
                            _line = (
                                f"[green]{_line}[/green]"
                                if _line.startswith("+")
                                else _line
                            )
                            _line = (
                                f"[red]{_line}[/red]"
                                if _line.startswith("-")
                                else _line
                            )
                            _line = (
                                f"[magenta]{_line}[/magenta]"
                                if _line.startswith("@")
                                else _line
                            )
                            self.stdout(_line, end="")
                            diff.append(line)

                        if (
                            patchcmd
                            and self._prompt("Accept this change?", "yN", answer) == "y"
                        ):
                            if self._patch(patchcmd, diff):
                                self._log(None, f"{path}: merged output\n", verbose)
                                os.remove(errpath)
                            else:
                                self._log(f"{path}: merge failed\n")

                return refout, postout, diff

            yield path, testwrapper

        if total[0] > 0:
            self._log("\n", None, verbose)
            self._log(
                (
                    f"# Ran [green]{total[0]}[/green] tests, "
                    f"[yellow]{skipped[0]}[/yellow] skipped, "
                    f"[red]{failed[0]}[/red] failed.\n"
                )
            )

    def _prompt(self, question, answers, auto=None):
        """Write a prompt to stdout and ask for answer in stdin.

        answers should be a string, with each character a single
        answer. An uppercase letter is considered the default answer.

        If an invalid answer is given, this asks again until it gets a
        valid one.

        If auto is set, the question is answered automatically with the
        specified value.
        """
        default = [c for c in answers if c.isupper()]
        while True:
            self.stdout(f"{question} [[blue]{answers}[/blue]] ", end="")
            if auto is not None:
                self.stdout(auto)
                return auto

            answer = sys.stdin.readline().strip().lower()
            if not answer and default:
                return default[0]
            elif answer and answer in answers.lower():
                return answer

    def _load_settings(self, argv):
        """Loads the settings from all layers and merges them.

        Layers (Config file > ENV vars > CLI arguments)"""
        argv = sys.argv[1:] if argv is None else argv
        argv.extend(_env_args("frysk"))
        args = self._argparser.parse_args(argv)
        self._color_mode(args.color)
        options = self._argparser.options

        try:
            configuration_settings = settings_from(
                load(
                    Path(self._expandpath(os.environ.get("fryskRC", ".fryskrc"))),
                    options,
                )
            )
        except ValueError as ex:
            raise _CliError(
                ExitCode.ERROR,
                "\n".join([f"{self._argparser.format_usage()}", f"frysk: error: {ex}"]),
            ) from ex

        argument_settings = settings_from(args)
        settings = merge_settings(argument_settings, configuration_settings)
        return settings

    def _setup(self):
        """
        Sets up a directory where tests will run (called `tmpdir`).
        Also set $TMPDIR, $TMP, and $TMP environment variables to a `tmp` dir inside the tmpdir.
        """
        self.tmpdir = os.environ["frysk_TEMP"] = tempfile.mkdtemp("", "frysk-tests-")
        self.tmpdir = Path(self.tmpdir)
        proc_tmp = self.tmpdir / "tmp"
        for name in ("TMPDIR", "TEMP", "TMP"):
            os.environ[name] = f"{proc_tmp}"

        os.mkdir(proc_tmp)

    def _cleanup(self):
        """
        A wrapper for #shutil.rmtree() that can try to remove write protection
        if removing fails, if enabled.
        """
        if self.tmpdir is None:
            return

        def on_rm_error(func, path, exc_info):
            os.chmod(path, stat.S_IWRITE)
            os.unlink(path)

        shutil.rmtree(self.tmpdir, onerror=on_rm_error)

    def main(self, argv=None):
        try:
            settings = self._load_settings(argv)
        except _CliError as ex:
            self.stderr(f"{ex}")
            return ex.exit_code

        conflict = _conflicts(settings)

        if conflict:
            arg1, arg2 = conflict
            self.stderr(f"options {arg1} and {arg2} are mutually exclusive")
            return ExitCode.ERROR

        shellcmd = which(settings.shell)
        if not shellcmd:
            self.stderr(f"shell not found: {settings.shell}")
            return ExitCode.ERROR
        shell = [shellcmd]
        if settings.shell_opts:
            shell += shlex.split(settings.shell_opts)

        patchcmd = None
        if settings.interactive:
            patchcmd = which("patch")
            if not patchcmd:
                self.stderr("patch(1) required for -i")
                return ExitCode.ERROR

        badpaths = [path for path in settings.tests if not path.exists()]
        if badpaths:
            self.stderr(f"no such file: {badpaths[0]}")
            return ExitCode.ERROR

        if settings.yes:
            answer = "y"
        elif settings.no:
            answer = "n"
        else:
            answer = None

        self._setup()  # sets self.tmpdir

        try:
            tests = runtests(
                settings.tests,
                self.tmpdir,
                shell,
                indent=settings.indent,
                cleanenv=not settings.preserve_env,
                debug=settings.debug,
                dos2unix=settings.dos2unix,
            )
            if not settings.debug:
                tests = self._runcli(
                    tests,
                    quiet=settings.quiet,
                    verbose=settings.verbose,
                    patchcmd=patchcmd,
                    answer=answer,
                )
                if settings.xunit_file is not None:
                    tests = runxunit(tests, settings.xunit_file)

            hastests = False
            failed = False
            for path, test in tests:
                hastests = True
                _, _, diff = test()
                if diff:
                    failed = True

            if not hastests:
                self.stderr("[red]no tests found[/red]")
                return ExitCode.ERROR

            return ExitCode.TEST_FAILED if failed else ExitCode.SUCCESS
        finally:
            if settings.keep_tmpdir:
                self.stdout(f"# Kept temporary directory: [blue]{self.tmpdir}[/blue]")
            else:
                self._cleanup()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import sys
from pathlib import Path

CURRENT_DIR = Path(__file__).parent.resolve()
sys.path.insert(0, f"{CURRENT_DIR.parent}")

# -- Project information -----------------------------------------------------

project = "frysk"
author = "Brodie Rao, Nicola Coretti & Contributors"
copyright = author
release = ""
# The full version, including alpha/beta/rc tags

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ["sphinx_copybutton", "sphinx_multiversion"]

# configure multiverse whitelist
smv_tag_whitelist = r"^.*$"  # Include all tags
smv_branch_whitelist = r"master"  # Include master

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "furo"
html_title = f"frysk {release}"
html_context = {
    "display_github": True,
    "github_user": "Nicoretti",
    "github_repo": project,
    "github_version": "master",
    "conf_py_path": "/docs/",
    "source_suffix": "rst",
}
html_sidebars = {
    "**": [
        "sidebar/scroll-start.html",
        "sidebar/brand.html",
        "sidebar/search.html",
        "sidebar/navigation.html",
        "versions.html",
        "sidebar/scroll-end.html",
    ]
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]



import argparse
from dataclasses import (
    dataclass,
    field,
)


@dataclass
class Settings:
    tests: list = field(default_factory=list)
    quiet: bool = None
    verbose: bool = None
    interactive: bool = None
    debug: bool = None
    yes: bool = None
    no: bool = None
    preserve_env: bool = None
    keep_tmpdir: bool = None
    shell: str = None
    shell_opts: list = field(default_factory=list)
    indent: int = None
    color: str = "auto"
    xunit_file: str = None
    dos2unix: bool = None


def settings_from(obj):
    supported_types = (
        argparse.Namespace,
        dict,
    )
    _class = type(obj)
    if not issubclass(_class, supported_types):
        raise TypeError(f"Can not construct settings from type: {type(obj).__name__}.")

    def from_namespace(ns):
        return Settings(**vars(ns))

    def from_dict(d):
        return Settings(**d)

    dispatcher = {
        f"{argparse.Namespace.__name__}": from_namespace,
        f"{dict.__name__}": from_dict,
    }

    return dispatcher[_class.__name__](obj)


def merge_settings(lhs, rhs):
    def items(d):
        d = vars(d)
        excludes = ["tests"]
        return ((k, v) for k, v in d.items() if k not in excludes)

    lhs_items = dict(items(lhs))
    rhs_items = dict(items(rhs))

    for name, value in rhs_items.items():
        value = value if value is not None else lhs_items.get(name, None)
        setattr(lhs, name, value)
    lhs.tests.extend(rhs.tests)
    lhs.shell_opts.extend(rhs.shell_opts)
    return lhs



"""Utilities for running individual tests"""

import itertools
import os
import re
import time
from contextlib import contextmanager
from pathlib import Path

from frysk.diff import (
    esc,
    glob,
    regex,
    unified_diff,
)
from frysk.process import (
    PIPE,
    STDOUT,
    execute,
)

__all__ = ["test", "testfile", "runtests"]

_SKIP = 80
_IS_ESCAPING_NEEDED = re.compile(rb"[\x00-\x09\x0b-\x1f\x7f-\xff]").search


def is_hidden(path):
    """Check if a path (file/dir) is hidden or not."""

    def _is_hidden(part):
        return (
            part.startswith(".")
            and not part == "."
            and not part.startswith("..")
            and not part.startswith("./")
        )

    return any(map(_is_hidden, path.parts))


def is_testfile(path):
    """Check if path is a valid frysk test file"""
    return path.is_file() and path.suffix == ".t" and not is_hidden(path)


def _escape(s):
    """Like the string-escape codec, but doesn't escape quotes"""
    escape_sub = re.compile(rb"[\x00-\x09\x0b-\x1f\\\x7f-\xff]").sub
    escape_map = dict((bytes([i]), rb"\x%02x" % i) for i in range(256))
    escape_map.update({b"\\": b"\\\\", b"\r": rb"\r", b"\t": rb"\t"})
    return escape_sub(lambda m: escape_map[m.group(0)], s[:-1]) + b" (esc)\n"


def _findtests(paths):
    """Yield tests in paths in sorted order"""

    paths = list(map(Path, paths))

    def is_test_dir(path):
        """Check if the path is a valid frysk test dir"""
        return path.is_dir() and not is_hidden(path)

    def remove_duplicates(path):
        """Stable duplication removal"""
        return list(dict.fromkeys(path))

    def collect(paths):
        """Collect all test files compliant with cram collection order"""
        for path in paths:
            if is_testfile(path):
                yield path
            if is_test_dir(path):
                yield from sorted((f for f in path.rglob("*.t") if is_testfile(f)))

    yield from remove_duplicates(collect(paths))


@contextmanager
def cwd(path):
    """Change the current working directory and restore it afterwards."""
    _cwd = os.getcwd()
    os.chdir(path)
    try:
        yield path
    finally:
        os.chdir(_cwd)


def test(
    lines,
    shell="/bin/sh",
    indent=2,
    testname=None,
    env=None,
    cleanenv=True,
    debug=False,
    dos2unix=False,
):
    r"""Run test lines and return input, output, and diff.

    This returns a 3-tuple containing the following:

        (list of lines in test, same list with actual output, diff)

    diff is a generator that yields the diff between the two lists.

    If a test exits with return code 80, the actual output is set to
    None and diff is set to [].

    Note that the TESTSHELL environment variable is available in the
    test (set to the specified shell). However, the TESTDIR and
    TESTFILE environment variables are not available. To run actual
    test files, see testfile().

    Example usage:

    >>> refout, postout, diff = test([b'  $ echo hi\n',
    ...                               b'  [a-z]{2} (re)\n'])
    >>> refout == [b'  $ echo hi\n', b'  [a-z]{2} (re)\n']
    True
    >>> postout == [b'  $ echo hi\n', b'  hi\n']
    True
    >>> bool(diff)
    False

    lines may also be a single bytes string:

    >>> refout, postout, diff = test(b'  $ echo hi\n  bye\n')
    >>> refout == [b'  $ echo hi\n', b'  bye\n']
    True
    >>> postout == [b'  $ echo hi\n', b'  hi\n']
    True
    >>> bool(diff)
    True
    >>> (b''.join(diff) ==
    ...  b'--- \n+++ \n@@ -1,2 +1,2 @@\n   $ echo hi\n-  bye\n+  hi\n')
    True

    :param lines: Test input
    :type lines: bytes or collections.Iterable[bytes]
    :param shell: Shell to run test in
    :type shell: bytes or str or list[bytes] or list[str]
    :param indent: Amount of indentation to use for shell commands
    :type indent: int
    :param testname: Optional test file name (used in diff output)
    :type testname: bytes or None
    :param env: Optional environment variables for the test shell
    :type env: dict or None
    :param cleanenv: Whether or not to sanitize the environment
    :type cleanenv: bool
    :param debug: Whether or not to run in debug mode (don't capture stdout)
    :type debug: bool
    :param dos2unix: Whether or not to convert all DOS/Windows line endings to UNIX
    :type debug: bool
    return: Input, output, and diff iterables
    :rtype: (list[bytes], list[bytes], collections.Iterable[bytes])
    """
    indent = b" " * indent
    cmdline = indent + b"$ "
    conline = indent + b"> "
    salt = b"frysk%.5f" % time.time()

    def create_environment(environment, shell, clean=False):
        _env = os.environ.copy() if environment is None else environment
        _env["TESTSHELL"] = shell
        if clean:
            _env.update({key: "C" for key in ["LANG", "LC_ALL", "LANGUAGE"]})
            _env["TZ"] = "GMT"
            _env["CDPATH"] = ""
            _env["COLUMNS"] = "80"
            _env["GREP_OPTIONS"] = ""
        return _env

    lines = lines.splitlines(True) if isinstance(lines, bytes) else lines
    shell = [shell] if isinstance(shell, (bytes, str)) else shell
    env = create_environment(env, shell[0], clean=cleanenv)

    if debug:
        return _debug(cmdline, conline, env, lines, shell)

    after = {}
    refout, postout = [], []
    i = pos = prepos = -1
    stdin = []
    for i, line in enumerate(lines):
        # Convert Windows style line endings to UNIX
        if dos2unix and line.endswith(b"\r\n"):
            line = line[:-2] + b"\n"
        if not line.endswith(b"\n"):
            line += b"\n"
        refout.append(line)
        if line.startswith(cmdline):
            after.setdefault(pos, []).append(line)
            prepos = pos
            pos = i
            stdin.append(b"echo %s %d $?\n" % (salt, i))
            stdin.append(line[len(cmdline) :])
        elif line.startswith(conline):
            after.setdefault(prepos, []).append(line)
            stdin.append(line[len(conline) :])
        elif not line.startswith(indent):
            after.setdefault(pos, []).append(line)
    stdin.append(b"echo %s %d $?\n" % (salt, i + 1))

    output, retcode = execute(
        shell + ["-"], stdin=b"".join(stdin), stdout=PIPE, stderr=STDOUT, env=env
    )
    if retcode == _SKIP:
        return refout, None, []

    pos = -1
    for i, line in enumerate(output[:-1].splitlines(True)):
        out, cmd = line, None
        if salt in line:
            out, cmd = line.split(salt, 1)

        if out:
            # Convert Windows style line endings to UNIX
            if dos2unix and out.endswith(b"\r\n"):
                out = out[:-2] + b"\n"
            if not out.endswith(b"\n"):
                out += b" (no-eol)\n"

            if _IS_ESCAPING_NEEDED(out):
                out = _escape(out)

            try:
                tmpdir = os.environ["TMPDIR"].encode()
            except KeyError:
                pass
            else:
                out = re.sub(re.escape(tmpdir), b"$TMPDIR", out)

            postout.append(indent + out)

        if cmd:
            ret = int(cmd.split()[1])
            if ret != 0:
                postout.append(indent + b"[%d]\n" % ret)
            postout += after.pop(pos, [])
            pos = int(cmd.split()[0])

    postout += after.pop(pos, [])

    if testname:
        diff_path = bytes(testname)
        error_path = bytes(Path(testname.parent, f"{testname.name}.err"))
    else:
        diff_path = error_path = b""

    diff = unified_diff(
        refout, postout, diff_path, error_path, matchers=[esc, glob, regex]
    )
    for line in diff:
        return refout, postout, itertools.chain([line], diff)
    return refout, postout, []


def _debug(cmdline, conline, env, lines, shell):
    stdin = []
    for line in lines:
        if not line.endswith(b"\n"):
            line += b"\n"
        if line.startswith(cmdline):
            stdin.append(line[len(cmdline) :])
        elif line.startswith(conline):
            stdin.append(line[len(conline) :])
    execute(shell + ["-"], stdin=b"".join(stdin), env=env)
    return [], [], []


def testfile(
    path,
    shell="/bin/sh",
    indent=2,
    env=None,
    cleanenv=True,
    debug=False,
    testname=None,
    dos2unix=False,
):
    """Run test at path and return input, output, and diff.

    This returns a 3-tuple containing the following:

        (list of lines in test, same list with actual output, diff)

    diff is a generator that yields the diff between the two lists.

    If a test exits with return code 80, the actual output is set to
    None and diff is set to [].

    Note that the TESTDIR, TESTFILE, and TESTSHELL environment
    variables are available to use in the test.

    :param path: Path to test file
    :type path: bytes or str
    :param shell: Shell to run test in
    :type shell: bytes or str or list[bytes] or list[str]
    :param indent: Amount of indentation to use for shell commands
    :type indent: int
    :param env: Optional environment variables for the test shell
    :type env: dict or None
    :param cleanenv: Whether or not to sanitize the environment
    :type cleanenv: bool
    :param debug: Whether or not to run in debug mode (don't capture stdout)
    :type debug: bool
    :param testname: Optional test file name (used in diff output)
    :type testname: bytes or None
    :return: Input, output, and diff iterables
    :rtype: (list[bytes], list[bytes], collections.Iterable[bytes])
    """
    with open(path, "rb") as f:
        abspath = path.resolve()
        env = env or os.environ.copy()
        env["TESTDIR"] = f"{abspath.parent}"
        env["TESTFILE"] = f"{abspath.name}"
        if testname is None:
            testname = path
        return test(
            f,
            shell,
            indent=indent,
            testname=testname,
            env=env,
            cleanenv=cleanenv,
            debug=debug,
            dos2unix=dos2unix,
        )


def runtests(
    paths, tmpdir, shell, indent=2, cleanenv=True, debug=False, dos2unix=False
):
    """Run tests and yield results.

    This yields a sequence of 2-tuples containing the following:

        (test path, test function)

    The test function, when called, runs the test in a temporary directory
    and returns a 3-tuple:

        (list of lines in the test, same list with actual output, diff)
    """
    basenames, seen = set(), set()
    tests = _findtests(paths)
    for i, path in enumerate(tests):
        abspath = path.resolve()
        if abspath in seen:
            continue
        seen.add(abspath)

        if not path.stat().st_size:
            yield path, lambda: (None, None, None)
            continue

        basename = path.name
        if basename in basenames:
            basename = f"{basename}-{i}"
        else:
            basenames.add(basename)

        def test():
            """Run test file"""
            testdir = tmpdir / basename
            os.mkdir(testdir)
            with cwd(testdir):
                return testfile(
                    abspath,
                    shell,
                    indent=indent,
                    cleanenv=cleanenv,
                    debug=debug,
                    testname=path,
                    dos2unix=dos2unix,
                )

        yield path, test



from inspect import cleandoc
from pathlib import Path

import nox
from nox import Session

BASEPATH = Path(__file__).parent.resolve()
DOCS = BASEPATH / "docs"
BUILD = DOCS / "_build"

nox.options.sessions = [
    "clean",
    "isort",
    "code_format",
    "pylint",
    "mypy",
    "unit",
    "integration",
    "coverage",
    "docs",
]


@nox.session(python=False)
def clean(_: Session) -> None:
    coverage_file = BASEPATH / ".coverage"
    coverage_file.unlink(missing_ok=True)


@nox.session(python=False)
def fix(session: Session) -> None:
    session.run("poetry", "run", "python", "-m", "isort", "-v", f"{BASEPATH}")
    session.run("poetry", "run", "python", "-m", "black", f"{BASEPATH}")


@nox.session(python=False)
def code_format(session: Session) -> None:
    session.run(
        "poetry",
        "run",
        "python",
        "-m",
        "black",
        "--check",
        "--diff",
        "--color",
        f"{BASEPATH}",
    )


@nox.session(python=False)
def isort(session: Session) -> None:
    session.run(
        "poetry", "run", "python", "-m", "isort", "-v", "--check", f"{BASEPATH}"
    )


@nox.session(python=False)
def pylint(session: Session) -> None:
    session.run("poetry", "run", "python", "-m", "pylint", f'{BASEPATH / "frysk"}')
    session.run("poetry", "run", "python", "-m", "pylint", f'{BASEPATH / "scripts"}')
    session.run(
        "poetry", "run", "python", "-m", "pylint", f'{BASEPATH / "pytest_frysk"}'
    )


@nox.session(python=False)
def unit(session: Session) -> None:
    session.env["COVERAGE"] = "coverage"
    session.env["COVERAGE_FILE"] = f'{BASEPATH / ".coverage"}'
    session.run(
        "poetry",
        "run",
        "coverage",
        "run",
        "-a",
        f'--rcfile={BASEPATH / "pyproject.toml"}',
        "-m",
        "pytest",
        "-v",
        "-p",
        "no:frysk",
        "--doctest-modules",
        f"{BASEPATH}",
    )


@nox.session(python=False)
@nox.parametrize("shell", ["dash", "bash", "zsh"])
def integration(session: Session, shell: str) -> None:
    session.env["TESTOPTS"] = f"--shell={shell}"
    session.env["COVERAGE"] = "coverage"
    session.env["COVERAGE_FILE"] = f'{BASEPATH / ".coverage"}'
    session.run(
        "poetry",
        "run",
        "coverage",
        "run",
        "-a",
        f'--rcfile={BASEPATH / "pyproject.toml"}',
        "-m",
        "frysk",
        f"--shell={shell}",
        f'{BASEPATH / "test" / "integration"}',
        external=True,
    )


@nox.session(python=False)
def mypy(session: Session) -> None:
    session.run(
        "poetry",
        "run",
        "mypy",
        "--strict",
        "--show-error-codes",
        "--pretty",
        "--show-column-numbers",
        "--show-error-context",
        "--scripts-are-modules",
    )


@nox.session(python=False)
def coverage(session: Session) -> None:
    session.env["COVERAGE"] = "coverage"
    session.env["COVERAGE_FILE"] = f'{BASEPATH / ".coverage"}'
    session.run("coverage", "report", "--fail-under=97")
    session.run("coverage", "lcov")


@nox.session(python=False)
def docs(session: Session) -> None:
    session.run("sphinx-build", f"{DOCS}", f"{BUILD}")


@nox.session(name="multi-version-docs", python=False)
def multi_version_docs(session: Session) -> None:
    session.run("sphinx-multiversion", f"{DOCS}", f"{BUILD}")
    with open(BUILD / "index.html", "w") as f:
        f.write(
            cleandoc(
                """
              <!DOCTYPE HTML>
              <html lang="en-US">
              <head>
                  <meta charset="UTF-8">
                  <meta http-equiv="refresh" content="0; url=master/index.html">
                  <title>Page Redirection</title>
              </head>
              <body>
                  <!-- Note: don't tell people to `click` the link, just tell them that it is a link. -->
                  If you are not redirected automatically, follow this <a href='master/index.html'>Documentation</a>.
              </body>
              """
            )
        )



""" Cli tool to calculate the md5 checksum for one or multiple files """
import sys
from argparse import (
    ArgumentParser,
    FileType,
)
from hashlib import md5
from typing import (
    Optional,
    Sequence,
)


def _create_parser() -> ArgumentParser:
    parser = ArgumentParser()
    parser.add_argument("files", metavar="FILE", type=FileType("rb"), nargs="+")
    return parser


def main(argv: Optional[Sequence[str]] = None) -> int:
    parser = _create_parser()
    args = parser.parse_args(argv)
    for file in args.files:
        hasher = md5()
        hasher.update(file.read())
        print(hasher.hexdigest())
    return 0


if __name__ == "__main__":
    sys.exit(main())
